from .apps import DesktopApp, Website, quick_website, quick_desktop_app, set_global_app, get_global_app
from .pages import Page, RedirectingPage
from .elements import Element, IFrameElement
from .structure import ToUIBlueprint
from . import exceptions

__version__ = "v3.0.2-beta"
