import sys
sys.path.append("..")
import os
os.chdir("../examples")

from examples import (example_1_simple_website,
                      example_2_simple_desktop_app,
                      example_3_updating_page,
                      example_4_function_with_arg,
                      example_5_user_variables,
                      example_6_quick_website,
                      example_7_quick_desktop_app,
                      advanced_example_1_toui_blueprint)

